import json
import pathlib
import typing as t

import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


class MyNode(zntrack.Node):
    deps_path: t.Union[list[t.Union[str, None]], None] = zntrack.deps_path()
    params_path: t.Union[list[t.Union[str, None]], None] = zntrack.params_path()
    outs_path: t.Union[list[t.Union[str, None]], None] = zntrack.outs_path()
    metrics_path: t.Union[list[t.Union[str, None]], None] = zntrack.metrics_path()
    plots_path: t.Union[list[t.Union[str, None]], None] = zntrack.plots_path()

    def run(self):
        pass


def test_x_path_none(proj_path):
    project = zntrack.Project()

    with project:
        _ = MyNode(
            deps_path=None,
            params_path=None,
            outs_path=None,
            metrics_path=None,
            plots_path=None,
        )
        _ = MyNode(
            deps_path=[None],
            params_path=[None],
            outs_path=[None],
            metrics_path=[None],
            plots_path=[None],
        )

    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "x_path_none.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "x_path_none.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "x_path_none.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()



"""Test passing callables as dependencies instead of properties or attributes."""

import contextlib
import typing as t

import zntrack

ContextManagerGenInt = t.Generator[int, None, None]


class NodeWithFunction(zntrack.Node):
    parameter: int = zntrack.params()

    def get_parameter(self) -> int:
        return self.parameter

    def run(self) -> None:
        pass


class NodeWithContextManager(zntrack.Node):
    parameter: int = zntrack.params()

    @contextlib.contextmanager
    def get_parameter(self) -> ContextManagerGenInt:
        yield self.parameter

    def run(self) -> None:
        pass


class NodeWithFunctionDeps(zntrack.Node):
    fn: t.Callable[[], int] = zntrack.deps()
    result: int = zntrack.outs()

    def run(self) -> None:
        self.result = self.fn()


class NodeWithContextManagerDeps(zntrack.Node):
    cm: t.Callable[[], t.ContextManager[int]] = zntrack.deps()
    result: int = zntrack.outs()

    def run(self) -> None:
        with self.cm() as value:
            self.result = value


def test_node_with_function(proj_path):
    project = zntrack.Project()

    with project:
        a = NodeWithFunction(parameter=1)
        b = NodeWithFunctionDeps(fn=a.get_parameter)

    project.repro()

    assert b.result == 1


def test_node_with_context_manager(proj_path):
    project = zntrack.Project()

    with project:
        a = NodeWithContextManager(parameter=1)
        b = NodeWithContextManagerDeps(cm=a.get_parameter)

    project.repro()

    assert b.result == 1



import dataclasses
import json
import logging
import typing as t

import znjson

from zntrack import config, converter
from zntrack.config import (
    FIELD_TYPE,
    PLUGIN_EMPTY_RETRUN_VALUE,
    ZNTRACK_FIELD_DUMP,
    ZNTRACK_FIELD_LOAD,
    ZNTRACK_FIELD_SUFFIX,
    FieldTypes,
)

# if t.TYPE_CHECKING:
from zntrack.plugins import ZnTrackPlugin
from zntrack.plugins.dvc_plugin.dvc import (
    deps_path_to_dvc,
    deps_to_dvc,
    metrics_path_to_dvc,
    metrics_to_dvc,
    outs_path_to_dvc,
    outs_to_dvc,
    params_path_to_dvc,
    plots_path_to_dvc,
    plots_to_dvc,
)
from zntrack.plugins.dvc_plugin.params import deps_to_params
from zntrack.utils.misc import (
    sort_and_deduplicate,
)

log = logging.getLogger(__name__)


@dataclasses.dataclass
class DVCPlugin(ZnTrackPlugin):
    def getter(self, field: dataclasses.Field) -> t.Any:
        getter = field.metadata.get(ZNTRACK_FIELD_LOAD)
        suffix = field.metadata.get(ZNTRACK_FIELD_SUFFIX)

        if getter is not None:
            if suffix is not None:
                return getter(self.node, field.name, suffix=suffix)
            return getter(self.node, field.name)
        return PLUGIN_EMPTY_RETRUN_VALUE

    def save(self, field: dataclasses.Field) -> None:
        dump_func = field.metadata.get(ZNTRACK_FIELD_DUMP)
        suffix = field.metadata.get(ZNTRACK_FIELD_SUFFIX)

        if dump_func is not None:
            if suffix is not None:
                dump_func(self.node, field.name, suffix=suffix)
            else:
                dump_func(self.node, field.name)

    def convert_to_params_yaml(self) -> dict | object:
        data = {}
        for field in dataclasses.fields(self.node):
            if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
                data[field.name] = getattr(self.node, field.name)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS:
                if (value := deps_to_params(self, field)) is not None:
                    data[field.name] = value

        if len(data) > 0:
            return data
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_dvc_yaml(self) -> dict | object:
        node_dict = converter.NodeConverter().encode(self.node)

        cmd = f"zntrack run {node_dict['module']}.{node_dict['cls']}"
        cmd += f" --name {node_dict['name']}"
        if hasattr(self.node, "_method"):
            cmd += f" --method {self.node._method}"
        stages = {
            "cmd": cmd,
            "metrics": [
                {
                    (self.node.nwd / "node-meta.json").as_posix(): {
                        "cache": config.ALWAYS_CACHE
                    }
                }
            ],
        }
        if self.node.always_changed:
            stages["always_changed"] = True
        plots = []

        for field in dataclasses.fields(self.node):
            if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
                stages.setdefault(FieldTypes.PARAMS.value, []).append(self.node.name)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS_PATH:
                content = params_path_to_dvc(self, field)
                stages.setdefault(FieldTypes.PARAMS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.OUTS_PATH:
                content = outs_path_to_dvc(self, field)
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS_PATH:
                content = plots_path_to_dvc(self, field)
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS_PATH:
                content = metrics_path_to_dvc(self, field)
                stages.setdefault(FieldTypes.METRICS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.OUTS:
                content = outs_to_dvc(self, field)
                stages.setdefault(FieldTypes.OUTS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS:
                outs_content, plots_content = plots_to_dvc(self, field)
                stages.setdefault(FieldTypes.OUTS.value, []).extend(outs_content)
                plots.extend(plots_content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS:
                content = metrics_to_dvc(self, field)
                stages.setdefault(FieldTypes.METRICS.value, []).extend(content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS:
                deps_content, params_content = deps_to_dvc(self, field)
                stages.setdefault(FieldTypes.DEPS.value, []).extend(deps_content)
                stages.setdefault(FieldTypes.PARAMS.value, []).extend(params_content)
            elif field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS_PATH:
                content = deps_path_to_dvc(self, field)
                stages.setdefault(FieldTypes.DEPS.value, []).extend(content)

        for key in list(stages):
            if key not in ["cmd", "always_changed"]:
                if len(stages[key]) == 0:
                    del stages[key]
                else:
                    stages[key] = sort_and_deduplicate(stages[key])

        return {"stages": stages, "plots": plots}

    def convert_to_zntrack_json(self, graph) -> dict | object:
        data = {
            "nwd": self.node.nwd,
        }
        for field in dataclasses.fields(self.node):
            if field.metadata.get(FIELD_TYPE) in [
                FieldTypes.PARAMS_PATH,
                FieldTypes.DEPS_PATH,
                FieldTypes.OUTS_PATH,
                FieldTypes.PLOTS_PATH,
                FieldTypes.METRICS_PATH,
                FieldTypes.DEPS,
            ]:
                data[field.name] = self.node.__dict__[field.name]

        data = znjson.dumps(
            data,
            indent=4,
            cls=znjson.ZnEncoder.from_converters(
                [
                    converter.ConnectionConverter,
                    converter.NodeConverter,
                    converter.CombinedConnectionsConverter,
                    znjson.converter.PathlibConverter,
                    converter.DVCImportPathConverter,
                    converter.DataclassConverter,
                ],
                add_default=False,
            ),
        )
        return json.loads(data)



import dataclasses
import json
import typing as t

import znflow
import znflow.handler
import znjson

from zntrack import converter
from zntrack.config import ZNTRACK_FILE_PATH, FieldTypes
from zntrack.fields.base import field
from zntrack.node import Node
from zntrack.utils.filesystem import resolve_state_file_path


def _deps_getter(self: "Node", name: str):
    zntrack_path = resolve_state_file_path(
        self.state.fs, self.state.path, ZNTRACK_FILE_PATH
    )

    with self.state.fs.open(zntrack_path) as f:
        content = json.load(f)[self.name][name]
        # TODO: Ensure deps are loaded from the correct revision
        try:
            content = znjson.loads(
                json.dumps(content),
                cls=znjson.ZnDecoder.from_converters(
                    [
                        converter.create_node_converter(
                            remote=self.state.remote or "",
                            rev=self.state.rev or "",
                            path=self.state.path,
                        ),
                        converter.ConnectionConverter,
                        converter.CombinedConnectionsConverter,
                        converter.DVCImportPathConverter,
                        converter.DataclassConverter,
                    ],
                    add_default=True,
                ),
            )
        except ModuleNotFoundError:
            # If external dataclass module can't be imported, return NOT_AVAILABLE
            # The enhanced NOT_AVAILABLE object will provide helpful errors when accessed
            from zntrack.config import NOT_AVAILABLE

            return NOT_AVAILABLE
        except AttributeError as e:
            # Only catch AttributeErrors related to missing module attributes
            if "module" in str(e).lower() or "attribute" in str(e).lower():
                from zntrack.config import NOT_AVAILABLE

                return NOT_AVAILABLE
            # Re-raise other AttributeErrors as they might indicate different issues
            raise
        if isinstance(content, converter.DataclassContainer):
            content = content.get_with_params(
                self.name,
                name,
                index=None,
                fs=self.state.fs,
                path=self.state.path,  # type: ignore
            )
        if isinstance(content, list):
            new_content = []
            idx = 0
            for val in content:
                if isinstance(val, converter.DataclassContainer):
                    new_content.append(
                        val.get_with_params(
                            self.name,
                            name,
                            idx,
                            fs=self.state.fs,
                            path=self.state.path,  # type: ignore
                        )
                    )
                    idx += 1  # index only runs over dataclasses
                else:
                    new_content.append(val)
            content = new_content

        content = znflow.handler.UpdateConnectors()(content)

        return content


def deps(default=dataclasses.MISSING, **kwargs) -> t.Any:
    """Define dependencies for a node.

    A Node dependency field can be used to pass data from one node to another.
    It can not be used to pass anything that is not a ``zntrack.Node`` or a ``dataclass``.

    Parameters
    ----------
    default : Any, optional
        Should not be set on the class level.

    Examples
    --------
    >>> import zntrack
    >>> class MyFirstNode(zntrack.Node):
    ...     outs: int = zntrack.outs()
    ...
    ...     def run(self) -> None:
    ...         self.outs = 42
    ...
    >>> class MySecondNode(zntrack.Node):
    ...     deps: int = zntrack.deps()
    ...
    ...     def run(self) -> None: ...
    ...
    >>> with zntrack.Project() as project:
    ...     node1 = MyFirstNode()
    ...     node2 = MySecondNode(deps=node1.outs)
    >>> project.build()
    """
    return field(
        default=default,
        load_fn=_deps_getter,
        field_type=FieldTypes.DEPS,
        **kwargs,
    )



import copy
import dataclasses
import pathlib

import znflow

from zntrack import converter
from zntrack.config import (
    FIELD_TYPE,
    ZNTRACK_CACHE,
    ZNTRACK_FIELD_SUFFIX,
    ZNTRACK_OPTION_PLOTS_CONFIG,
    FieldTypes,
)
from zntrack.node import Node
from zntrack.utils.misc import (
    RunDVCImportPathHandler,
    get_attr_always_list,
)
from zntrack.utils.node_wd import NWDReplaceHandler, nwd


def _handle_nwd_list(self, field):
    """Helper function to process lists of paths with NWD replacement."""
    if getattr(self.node, field.name) is None:
        return []
    nwd_handler = NWDReplaceHandler()
    return nwd_handler(get_attr_always_list(self.node, field.name), nwd=self.node.nwd)


def _paths_to_dvc_list(
    paths: list[str], as_dict: bool = False, cache: bool = True
) -> list[str] | list[dict]:
    """Helper function to convert lists of paths to DVC format."""
    content = [pathlib.Path(x).as_posix() for x in paths if x is not None]
    if not cache:
        return [{c: {"cache": False}} for c in content]
    return content if not as_dict else [{c: None} for c in content]


def params_path_to_dvc(self, field) -> list[str] | list[dict]:
    """Convert params_path field to DVC params format."""
    paths = _handle_nwd_list(self, field)
    return _paths_to_dvc_list(paths, as_dict=True)


def outs_path_to_dvc(self, field) -> list[str] | list[dict]:
    """Convert outs_path field to DVC outs format."""
    if getattr(self.node, field.name) == nwd:
        raise ValueError(
            "Can not use 'zntrack.nwd' directly as an output path. "
            "Please use 'zntrack.nwd / <path/file>' instead."
        )
    paths = _handle_nwd_list(self, field)
    return _paths_to_dvc_list(paths, cache=field.metadata.get(ZNTRACK_CACHE, True))


def plots_path_to_dvc(self, field) -> list[str] | list[dict]:
    """Convert plots_path field to DVC outs format."""
    paths = _handle_nwd_list(self, field)
    return _paths_to_dvc_list(paths, cache=field.metadata.get(ZNTRACK_CACHE, True))


def metrics_path_to_dvc(self, field) -> list[str] | list[dict]:
    """Convert metrics_path field to DVC metrics format."""
    paths = _handle_nwd_list(self, field)
    return _paths_to_dvc_list(paths, cache=field.metadata.get(ZNTRACK_CACHE, True))


def outs_to_dvc(self, field) -> list[str] | list[dict]:
    """Convert outs field to DVC outs format."""
    suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
    content = [(self.node.nwd / field.name).with_suffix(suffix).as_posix()]
    if field.metadata.get(ZNTRACK_CACHE) is False:
        return [{c: {"cache": False}} for c in content]
    return content


def metrics_to_dvc(self, field) -> list[str] | list[dict]:
    """Convert metrics field to DVC metrics format."""
    suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
    content = [(self.node.nwd / field.name).with_suffix(suffix).as_posix()]
    if field.metadata.get(ZNTRACK_CACHE) is False:
        return [{c: {"cache": False}} for c in content]
    return content


def deps_path_to_dvc(self, field) -> list[str]:
    """Convert deps_path field to DVC deps format."""
    if getattr(self.node, field.name) is None:
        return []
    content = [
        pathlib.Path(c).as_posix()
        for c in get_attr_always_list(self.node, field.name)
        if c is not None
    ]
    RunDVCImportPathHandler()(self.node.__dict__.get(field.name))
    return content


def deps_to_dvc(self, field) -> tuple[list[str], list[str]]:
    """Convert deps field to DVC deps and params format."""
    deps_content = []
    params_content = []
    if getattr(self.node, field.name) is None:
        return deps_content, params_content

    nwd_handler = NWDReplaceHandler()
    content = get_attr_always_list(self.node, field.name)
    paths = []

    for con in content:
        if isinstance(con, znflow.Connection):
            if con.item is not None:
                raise NotImplementedError(
                    "znflow.Connection getitem is not supported yet."
                )
            paths.extend(converter.node_to_output_paths(con.instance, con.attribute))
        elif isinstance(con, znflow.CombinedConnections):
            for _con in con.connections:
                if _con.item is not None:
                    raise NotImplementedError(
                        "znflow.Connection getitem is not supported yet."
                    )
                paths.extend(
                    converter.node_to_output_paths(_con.instance, _con.attribute)
                )
        elif dataclasses.is_dataclass(con) and not isinstance(con, Node):
            for sub_field in dataclasses.fields(con):
                field_type = sub_field.metadata.get(FIELD_TYPE)
                if field_type == FieldTypes.PARAMS_PATH:
                    content = nwd_handler(
                        get_attr_always_list(con, sub_field.name),
                        nwd=self.node.nwd,
                    )
                    param_paths = _paths_to_dvc_list(content, as_dict=True)
                    if param_paths:
                        params_content.extend(param_paths)
                elif field_type == FieldTypes.DEPS_PATH:
                    dep_paths = [
                        pathlib.Path(c).as_posix()
                        for c in get_attr_always_list(con, sub_field.name)
                        if c is not None
                    ]
                    if dep_paths:
                        deps_content.extend(dep_paths)
            params_content.append(self.node.name)  # Add node name to params
        else:
            raise ValueError(f"Unsupported type: {type(con)}")

    return deps_content + paths, params_content


def plots_to_dvc(self, field) -> tuple[list, list[dict]]:
    """Convert plots field to DVC outs and plots format."""
    outs_content = []
    plots_content = []
    suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
    file_path = (self.node.nwd / field.name).with_suffix(suffix).as_posix()
    outs_content.append(file_path)
    if field.metadata.get(ZNTRACK_CACHE) is False:
        outs_content = [{c: {"cache": False}} for c in outs_content]

    plots_config = field.metadata.get(ZNTRACK_OPTION_PLOTS_CONFIG)
    if plots_config:
        plots_config = plots_config.copy()
        if "x" not in plots_config or "y" not in plots_config:
            raise ValueError("Both 'x' and 'y' must be specified in the plots_config.")
        if "x" in plots_config:
            plots_config["x"] = {file_path: plots_config["x"]}
        if isinstance(plots_config["y"], list):
            for idx, y in enumerate(plots_config["y"]):
                cfg = copy.deepcopy(plots_config)
                cfg["y"] = {file_path: y}
                plots_content.append({f"{self.node.name}_{field.name}_{idx}": cfg})
        elif "y" in plots_config:
            plots_config["y"] = {file_path: plots_config["y"]}
            plots_content.append({f"{self.node.name}_{field.name}": plots_config})

    return outs_content, plots_content



import os
import pathlib

import pytest
from dvc.api import DVCFileSystem
from git import Repo

import zntrack.examples


def test_subrepo(proj_path):
    """Test subrepo functionality"""
    directory = pathlib.Path("subrepo")
    directory.mkdir(parents=True, exist_ok=True)
    os.chdir(directory)

    project = zntrack.Project()
    with project:
        node = zntrack.examples.ParamsToOuts(
            params={"param1": 1, "param2": 2},
        )
    project.repro()

    os.chdir(proj_path)

    node_loaded = zntrack.from_rev(
        "subrepo/dvc.yaml:ParamsToOuts",
    )
    assert node_loaded.params == {"param1": 1, "param2": 2}
    assert node_loaded.outs == {"param1": 1, "param2": 2}

    # now test with DVCFs
    node_loaded = zntrack.from_rev("subrepo/dvc.yaml:ParamsToOuts", fs=DVCFileSystem())
    assert node_loaded.params == {"param1": 1, "param2": 2}
    assert node_loaded.outs == {"param1": 1, "param2": 2}

    # make a commit and change things

    repo = Repo()
    repo.git.add(all=True)
    repo.index.commit("Initial commit")

    os.chdir(directory)
    node.params = {"param1": 3, "param2": 4}
    project.repro()
    os.chdir(proj_path)

    node_loaded = zntrack.from_rev(
        "subrepo/dvc.yaml:ParamsToOuts",
    )
    assert node_loaded.params == {"param1": 3, "param2": 4}
    assert node_loaded.outs == {"param1": 3, "param2": 4}

    node_loaded = zntrack.from_rev("subrepo/dvc.yaml:ParamsToOuts", rev="HEAD")
    assert node_loaded.params == {"param1": 1, "param2": 2}
    assert node_loaded.outs == {"param1": 1, "param2": 2}

    # now try loading the node using an absolute path
    with pytest.raises(FileNotFoundError):
        # DVCFileSystem struggles with absolute paths
        node_loaded = zntrack.from_rev(
            "subrepo/dvc.yaml:ParamsToOuts",
            remote=proj_path.resolve().as_posix(),
        )


def test_subrepo_external_node(proj_path):
    """Test subrepo functionality with external dataclasses."""
    # Create external node module file that can be imported
    external_module_content = """
from dataclasses import dataclass

@dataclass
class ExampleExternalNode:
    parameter: str
"""

    external_module_path = proj_path / "external_nodes.py"
    with open(external_module_path, "w") as f:
        f.write(external_module_content)

    # Import the external node from the file
    import importlib.util
    import sys

    # Store original sys.modules state for cleanup
    original_modules = sys.modules.copy()

    try:
        spec = importlib.util.spec_from_file_location(
            "external_nodes", external_module_path
        )
        external_nodes = importlib.util.module_from_spec(spec)
        sys.modules["external_nodes"] = external_nodes
        spec.loader.exec_module(external_nodes)

        example_external_node = external_nodes.ExampleExternalNode

        directory = pathlib.Path("subrepo")
        directory.mkdir(parents=True, exist_ok=True)
        os.chdir(directory)

        project = zntrack.Project()

        ext_node = example_external_node(parameter="Lorem Ipsum")

        with project:
            node = zntrack.examples.OptionalDeps(
                value=ext_node,
            )
        project.build()

        os.chdir(proj_path)

        assert node.value.parameter == "Lorem Ipsum"

        # Test: External dependencies behavior with nested repos
        # Remove the module from sys.modules to simulate it not being available
        if "external_nodes" in sys.modules:
            del sys.modules["external_nodes"]

        # Load node - should raise error when external module is not available
        node = zntrack.from_rev("subrepo/dvc.yaml:OptionalDeps")
        # The value itself will be NOT_AVAILABLE, but accessing attributes should raise
        assert node.value is zntrack.NOT_AVAILABLE

        # Accessing attributes on NOT_AVAILABLE should raise helpful error
        with pytest.raises(
            ModuleNotFoundError, match="Cannot access attribute.*external dependency"
        ):
            _ = node.value.parameter

    finally:
        # Restore original sys.modules state to avoid affecting other tests
        sys.modules.clear()
        sys.modules.update(original_modules)



from pathlib import Path

import zntrack.examples


def test_outs_path_to_deps_path(proj_path):
    (proj_path / "output.txt").write_text("15")

    with zntrack.Project() as proj:
        a = zntrack.examples.WriteDVCOuts(params=10)
        b = zntrack.examples.ReadFile(path=a.outs)
        #
        c = zntrack.examples.ReadFile(path=Path("output.txt"))

    proj.repro()

    assert a.outs == Path("nodes/WriteDVCOuts/output.txt")
    assert b.path == Path("nodes/WriteDVCOuts/output.txt")
    assert b.content == "10"
    # control
    assert c.path == Path("output.txt")
    assert c.content == "15"



import typing as t
from multiprocessing import Process, Queue

from dvc.api import DVCFileSystem
from dvc.stage.serialize import to_single_stage_lockfile


def get_stage_lock(name: str, queue: Queue) -> None:
    fs = DVCFileSystem()
    stage = next(iter(fs.repo.stage.collect(name)))
    stage.save_deps(allow_missing=False)
    result = to_single_stage_lockfile(stage)
    result = {k: v for k, v in result.items() if k in ["cmd", "deps", "params"]}
    queue.put(result)  # Send the result back to the main process


def mp_start_stage_lock(name: str) -> t.Tuple[Queue, Process]:
    queue = Queue()
    p = Process(target=get_stage_lock, args=(name, queue))
    p.start()
    return queue, p


def mp_join_stage_lock(queue: Queue, p: Process) -> dict:
    p.join()
    return queue.get()



import contextlib
import dataclasses
import datetime
import json
import logging
import pathlib
import typing as t
import uuid
import warnings

import dvc.api
import typing_extensions as ty_ex
import znfields
import znflow
from dvc.stage.exceptions import InvalidStageName
from dvc.stage.utils import is_valid_name
from fsspec.implementations.local import LocalFileSystem

from zntrack.group import Group
from zntrack.state import NodeStatus
from zntrack.utils.misc import get_plugins_from_env, nwd_to_name

from .config import (
    FIELD_TYPE,
    NOT_AVAILABLE,
    NWD_PATH,
    ZNTRACK_LAZY_VALUE,
    FieldTypes,
    NodeStatusEnum,
)

try:
    from typing import dataclass_transform
except ImportError:
    from typing_extensions import dataclass_transform

T = t.TypeVar("T", bound="Node")

log = logging.getLogger(__name__)


def _name_setter(self, attr_name: str, value: str) -> None:
    """
    Validates and sets the node name, ensuring uniqueness within the graph.

    Parameters
    ----------
    attr_name : str
        The attribute name to be set.
    value : str
        The desired node name.

    Raises
    ------
    AttributeError
        If the node name is already set and cannot be changed.
    InvalidStageName
        If the given name is invalid.
    ValueError
        If a node with the same name already exists.
    """
    if attr_name in self.__dict__:
        raise AttributeError("Node name cannot be changed.")

    if value is None:
        return

    if not is_valid_name(value):
        raise InvalidStageName

    if "_" in value:
        warnings.warn(
            "Node name should not contain '_'."
            " This character is used for defining groups."
        )
    self.__dict__[attr_name] = value  # only used to check if the name has been set once

    graph = znflow.get_graph()

    # in case the name is set outside the graph
    # (no groups can be active)
    nwd = NWD_PATH / value

    if graph is not znflow.empty_graph:
        name = self.__class__.__name__
        if graph.active_group is None:
            if graph.node_name_counter.get(name, 0) > 0:
                graph.node_name_counter[name] -= 1

            if value in graph.node_name_counter:
                raise ValueError(f"A node with the name '{value}' already exists.")
            graph.node_name_counter[value] = 1
        else:
            group_path = "/".join(graph.active_group.names)
            grp_and_name = f"{group_path}/{name}"

            if graph.node_name_counter.get(grp_and_name, 0) > 0:
                graph.node_name_counter[grp_and_name] -= 1

            node_name = f"{group_path}_{value}"
            if node_name in graph.node_name_counter:
                raise ValueError(f"A node with the name '{node_name}' already exists.")
            graph.node_name_counter[node_name] = 1
            nwd = NWD_PATH / group_path / value

    self.__dict__["nwd"] = nwd


def _name_getter(self, attr_name: str) -> str:
    """Retrieve the name of a node based on the current graph context.

    Parameter
    ---------
        attr_name (str): The attribute name to retrieve.

    Returns
    -------
        str: The resolved node name.

    """

    if self.__dict__.get("nwd") is not None:
        # can not use self.nwd in case of `tmp_path`
        return nwd_to_name(self.__dict__["nwd"])
    else:
        return self.__class__.__name__


@dataclass_transform()
@dataclasses.dataclass(kw_only=True)
class Node(znflow.Node, znfields.Base):
    """A Node."""

    name: str | None = znfields.field(
        default=None, getter=_name_getter, setter=_name_setter
    )
    always_changed: bool = dataclasses.field(default=False, repr=False)

    _protected_: set = dataclasses.field(
        default_factory=lambda: set(znflow.Node._protected_) | {"nwd", "name", "state"},
        init=False,
        repr=False,
    )

    def __post_init__(self):
        if self.name is None:
            # automatic node names expects the name to be None when
            # exiting the graph context.
            if not znflow.get_graph() is not znflow.empty_graph:
                self.name = self.__class__.__name__

        for field in dataclasses.fields(self):
            # X_Path should be resolved instead of passing
            #  a connection. They are known at runtime.
            if field.metadata.get(FIELD_TYPE, None) in [
                FieldTypes.PARAMS_PATH,
                FieldTypes.DEPS_PATH,
                FieldTypes.OUTS_PATH,
                FieldTypes.PLOTS_PATH,
                FieldTypes.METRICS_PATH,
            ]:
                self._protected_.add(field.name)

    def _post_load_(self):
        """Called after `from_rev` is called."""
        raise NotImplementedError

    def run(self):
        raise NotImplementedError

    def save(self):
        for plugin in self.state.plugins.values():
            with plugin:
                for field in dataclasses.fields(self):
                    value = getattr(self, field.name)
                    if any(value is x for x in [ZNTRACK_LAZY_VALUE, NOT_AVAILABLE]):
                        raise ValueError(
                            f"Field '{field.name}' is not set."
                            " Please set it before saving."
                        )
                    try:
                        plugin.save(field)
                    except Exception as err:  # noqa: E722
                        if plugin._continue_on_error_:
                            warnings.warn(
                                f"Plugin {plugin.__class__.__name__} failed to"
                                f" save field {field.name}."
                            )
                        else:
                            raise err

        _ = self.state
        self.__dict__["state"]["state"] = NodeStatusEnum.FINISHED

    def __init_subclass__(cls):
        return dataclasses.dataclass(cls, kw_only=True)

    @property
    def nwd(self) -> pathlib.Path:
        return self.state.nwd

    @classmethod
    def from_rev(
        cls: t.Type[T],
        name: str | None = None,
        remote: str | None = None,
        rev: str | None = None,
        running: bool = False,
        lazy_evaluation: bool = True,
        path: str | None | pathlib.Path = None,
        fs: dvc.api.DVCFileSystem | None = None,
        **kwargs,
    ) -> T:
        if name is None:
            name = cls.__name__
        if path is not None:
            path = pathlib.Path(path)
        else:
            path = pathlib.Path()
        lazy_values = {}
        for field in dataclasses.fields(cls):
            # check if the field is in the init
            if field.init:
                lazy_values[field.name] = ZNTRACK_LAZY_VALUE

        lazy_values["name"] = name
        lazy_values["always_changed"] = None  # TODO: read the state from dvc.yaml
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message=".*should not contain '_'*", category=UserWarning
            )
            instance = cls(**lazy_values)
        if remote is not None or rev is not None:
            if fs is None:
                _fs = dvc.api.DVCFileSystem(url=remote, rev=rev)
            else:
                _fs = fs
            with _fs.open(path / "zntrack.json") as f:
                conf = json.loads(f.read())
                nwd = pathlib.Path(conf[name]["nwd"]["value"])
        else:
            try:
                with open(path / "zntrack.json") as f:
                    conf = json.load(f)
                    nwd = pathlib.Path(conf[name]["nwd"]["value"])
            except FileNotFoundError:
                # from_rev is called before a graph is built
                nwd = NWD_PATH / name
        instance.__dict__["nwd"] = nwd

        # TODO: check if the node is finished or not.
        if fs is None:
            if remote is not None or rev is not None:
                fs = dvc.api.DVCFileSystem(url=remote, rev=rev)
            else:
                fs = LocalFileSystem()
        instance.__dict__["state"] = NodeStatus(
            remote=remote,
            rev=rev,
            state=NodeStatusEnum.RUNNING if running else NodeStatusEnum.FINISHED,
            lazy_evaluation=lazy_evaluation,
            group=Group.from_nwd(instance.nwd),
            path=path,
            fs=fs,
        ).to_dict()
        instance.__dict__["state"]["plugins"] = get_plugins_from_env(instance)

        with contextlib.suppress(FileNotFoundError):
            # need to update run_count after the state is set
            # TODO: do we want to set the UUID as well?
            # TODO: test that run_count is correct, when using from_rev from another
            #  commit
            with instance.state.fs.open(path / instance.nwd / "node-meta.json") as f:
                content = json.load(f)
                run_count = content.get("run_count", 0)
                run_time = content.get("run_time", 0)
                lockfile = content.get("lockfile", None)
                if node_uuid := content.get("uuid", None):
                    instance._uuid = uuid.UUID(node_uuid)
                instance.__dict__["state"]["run_count"] = run_count
                instance.__dict__["state"]["run_time"] = datetime.timedelta(
                    seconds=run_time
                )
                instance.__dict__["state"]["lockfile"] = lockfile
        if not instance.state.lazy_evaluation:
            for field in dataclasses.fields(cls):
                _ = getattr(instance, field.name)

        instance._external_ = True
        if not running and hasattr(instance, "_post_load_"):
            with contextlib.suppress(NotImplementedError):
                instance._post_load_()

        return instance

    @property
    def state(self) -> NodeStatus:
        if "state" not in self.__dict__:
            self.__dict__["state"] = NodeStatus().to_dict()
            self.__dict__["state"]["plugins"] = get_plugins_from_env(self)

        return NodeStatus(**self.__dict__["state"], node=self)

    @ty_ex.deprecated("loading is handled automatically via lazy evaluation")
    def load(self):
        pass



import importlib
import pathlib
import sys

import dvc.api
import git
from dvc.scm import SCMError
from dvc.stage.exceptions import StageFileDoesNotExistError


def from_rev(
    name: str,
    remote: str | None = None,
    rev: str | None = None,
    path: str | None = None,
    fs: dvc.api.DVCFileSystem | None = None,
):
    """Load a ZnTrack Node.

    Load an instance of any ZnTrack Node, given its name.

    Arguments
    ---------
    name : str
        The name of the ZnTrack Node to load.
        If multiple ``dvc.yaml`` files are present, the name should be
        specified as `path/to/dvc.yaml:NodeName`.
    remote : str, optional
        The remote URL where the DVC repository is located.
        If not provided, the current working directory will be used.
        Can be a local path or a remote URL (e.g., git repository).
    rev : str, optional
        The revision (commit hash, branch name, or tag) to load the Node from.
        If not provided, the current WORKSPACE revision will be used.
    fs: dvc.api.DVCFileSystem, optional
        A DVCFileSystem instance to use for accessing the DVC repository.
        If not provided, a new DVCFileSystem will be created using the `remote` and `rev`.
    """
    if path is not None:
        raise NotImplementedError
    if fs is None:
        fs = dvc.api.DVCFileSystem(url=remote, rev=rev)
    else:
        if remote is not None:
            raise ValueError(
                "If 'fs' is provided, 'remote' should be None. "
                "The remote is already specified in the DVCFileSystem."
            )
        if rev is not None:
            raise ValueError(
                "If 'fs' is provided, 'rev' should be None. "
                "The revision is already specified in the DVCFileSystem."
            )
        # get remote and rev from the fs
        remote = fs.repo.url
        try:
            rev = fs.repo.get_rev()
            # check if rev is the same as HEAD, then set to None
            try:
                if rev == git.Repo(fs.repo.root_dir).head.commit.hexsha:
                    rev = None
            except (git.exc.InvalidGitRepositoryError, git.exc.NoSuchPathError):
                # If we can't access the local git repo, just use the rev as-is
                pass
        except SCMError:
            rev = None
    try:
        stage = fs.repo.stage.collect(target=name)[0]
    except StageFileDoesNotExistError:
        raise ValueError(f"Stage {name} not found in {fs.repo}")

    try:
        cmd = stage.cmd
        name = stage.name
        path = pathlib.Path(stage.path).parent
    except AttributeError:
        raise ValueError("Stage is not a ZnTrack pipeline stage.")

    # cmd will be "zntrack run module.name --name ..."
    # and we need the module.name and --name part
    run_str = cmd.split()[2]
    name = cmd.split()[4]

    package_and_module, cls_name = run_str.rsplit(".", 1)

    sys.path.append(pathlib.Path.cwd().as_posix())

    # If we have a filesystem with a local repo, add it to Python path
    if fs is not None and hasattr(fs, "repo") and fs.repo is not None:
        repo_root = pathlib.Path(fs.repo.root_dir)
        if repo_root.exists():
            repo_root_str = str(repo_root)
            if repo_root_str not in sys.path:
                sys.path.insert(0, repo_root_str)

    try:
        module = importlib.import_module(package_and_module)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"The node depends on package '{package_and_module}' which is not installed "
            f"in the current environment. You can install it via:\n"
            f"  pip install {package_and_module}\n"
            f"Or if it's available from the remote repository:\n"
            f"  pip install git+{remote if remote else 'REMOTE_URL'}"
        )

    cls = getattr(module, cls_name)
    return cls.from_rev(name, remote=remote, rev=rev, path=path, fs=fs)



import subprocess

import zntrack.examples
from zntrack.utils.state import get_node_status


def test_get_node_status(proj_path):
    project = zntrack.Project()

    with project:
        a = zntrack.examples.ParamsToOuts(params=42)
        b = zntrack.examples.ParamsToOuts(params=43)
        c = zntrack.examples.AddNodeNumbers(numbers=[a, b])

    project.build()

    subprocess.check_call(["dvc", "repro", a.name])

    assert get_node_status(a.name, remote=None, rev=None) is False
    assert get_node_status(b.name, remote=None, rev=None) is True
    assert get_node_status(c.name, remote=None, rev=None) is True
    assert get_node_status("non_existing_node", remote=None, rev=None) is None

    subprocess.check_call(["dvc", "repro"])

    assert get_node_status(a.name, remote=None, rev=None) is False
    assert get_node_status(b.name, remote=None, rev=None) is False
    assert get_node_status(c.name, remote=None, rev=None) is False

    a.params = 44
    project.build()

    assert get_node_status(a.name, remote=None, rev=None) is True
    assert get_node_status(b.name, remote=None, rev=None) is False
    assert (
        get_node_status(c.name, remote=None, rev=None) is False
    )  # DVC the deps have not been updated, so this still assumes it is True

    subprocess.check_call(["dvc", "repro", a.name])
    assert get_node_status(a.name, remote=None, rev=None) is False
    assert get_node_status(b.name, remote=None, rev=None) is False
    assert (
        get_node_status(c.name, remote=None, rev=None) is True
    )  # Now the deps have been updated, so this is True again



import dataclasses
import typing as t

import yaml

from zntrack.config import PARAMS_FILE_PATH, FieldTypes
from zntrack.fields.base import field
from zntrack.node import Node
from zntrack.utils.filesystem import resolve_state_file_path

_T = t.TypeVar("_T")


def _params_getter(self: "Node", name: str):
    params_path = resolve_state_file_path(
        self.state.fs, self.state.path, PARAMS_FILE_PATH
    )

    with self.state.fs.open(params_path) as f:
        return yaml.safe_load(f)[self.name][name]


# Overloads for type checking
@t.overload
def params(default: _T, **kwargs) -> _T: ...


@t.overload
def params(*, default_factory: t.Callable[[], _T], **kwargs) -> _T: ...


def params(
    default=dataclasses.MISSING, *, default_factory=dataclasses.MISSING, **kwargs
) -> t.Any:
    """ZnTrack parameter field.

    A field to define a parameter for a ZnTrack node.

    Parameters
    ----------
    default : dict|int|str|float|list|None, optional
        Set a default parameter value.

    default_factory : callable, optional
        A callable that returns the default value.
        Should be used instead of `default` if the default value is mutable.

    Examples
    --------

    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     param: int = zntrack.params(default=42)
    ...
    ...     def run(self) -> None: ...
    ...
    >>> a = MyNode()
    >>> a.param
    42
    >>> b = MyNode(param=43)
    >>> b.param
    43

    """
    # TODO: check types, do not allow e.g. connections
    #  or anything that can not be serialized
    return field(
        default=default,
        default_factory=default_factory,
        field_type=FieldTypes.PARAMS,
        load_fn=_params_getter,
        suffix=None,
        cache=kwargs.pop("cache", True),
        independent=kwargs.pop("independent", False),
        **kwargs,
    )



import json
from pathlib import Path

import dvc.api
import dvc.fs
from dvc.stage.serialize import to_single_stage_lockfile


def get_node_status(
    addressing: str,
    remote: str | None,
    rev: str | None,
    fs: dvc.api.DVCFileSystem | None = None,
) -> bool | None:
    """Check if a node has changed since the last DVC run.

    Returns:
        - True if the node has changed
        - False if the node has not changed
        - None if the node does not exist or is not a zntrack node
    """

    # TODO: this should check the node status of all dependencies, and if
    # any of them has changed, the value should be set to None / unknown.
    if fs is None:
        fs = dvc.fs.DVCFileSystem(
            repo=remote,
            rev=rev,
        )
    try:
        stage = next(iter(fs.repo.stage.collect(addressing)))
    except Exception:
        return None
    stage.save_deps(allow_missing=True)
    dvc_lock = to_single_stage_lockfile(stage)
    dvc_lock = {k: v for k, v in dvc_lock.items() if k in ["cmd", "deps", "params"]}

    try:
        zntrack_meta = json.loads(
            fs.read_text(Path(stage.path_in_repo).parent / "zntrack.json"),
        )
    except FileNotFoundError:
        return None

    try:
        nwd = Path(zntrack_meta[stage.name]["nwd"]["value"])
    except KeyError:
        return None

    nwd_in_repo = Path(stage.path_in_repo).parent / nwd
    try:
        node_meta = json.loads(fs.read_text(nwd_in_repo / "node-meta.json"))
    except FileNotFoundError:
        return True

    try:
        node_lock = node_meta["lockfile"]
    except KeyError:
        return None  # old zntrack version

    # Compare dvc_lock with node_lock
    if dvc_lock == node_lock:
        return False
    else:
        return True



"""Filesystem utilities for ZnTrack."""

import pathlib
from typing import Union


def resolve_dvc_path(
    fs, state_path: pathlib.Path, target_path: Union[str, pathlib.Path]
) -> Union[str, pathlib.Path]:
    """Resolve a path for DVCFileSystem, handling relative/absolute path conversion.

    DVCFileSystem expects paths relative to the repository root, but ZnTrack often
    works with absolute paths. This function handles the conversion.

    Args:
        fs: The filesystem object (DVCFileSystem or LocalFileSystem)
        state_path: The state path of the node
        target_path: The target path to resolve (can be str or Path)

    Returns:
        Resolved path suitable for the filesystem (str for DVCFileSystem, Path for local)
    """
    # For non-DVC filesystems, return the original path
    if not hasattr(fs, "repo") or not getattr(fs, "repo", None):
        return target_path

    # For DVCFileSystem, convert to relative path
    repo_root = pathlib.Path(fs.repo.root_dir)
    target_path = pathlib.Path(target_path)

    # If target_path is already relative, combine with state_path if needed
    if not target_path.is_absolute():
        if state_path.is_absolute():
            # Make state_path relative to repo_root first
            try:
                relative_state = state_path.relative_to(repo_root)
                final_path = relative_state / target_path
            except ValueError:
                # If state_path can't be made relative, use absolute target_path
                final_path = target_path
        else:
            # Both are relative
            final_path = state_path / target_path
    else:
        # target_path is absolute, make it relative to repo_root
        try:
            final_path = target_path.relative_to(repo_root)
        except ValueError:
            # If target_path can't be made relative to repo_root, use as-is
            final_path = target_path

    return str(final_path)


def resolve_state_file_path(
    fs, state_path: pathlib.Path, filename: str, validate_exists: bool = False
) -> Union[str, pathlib.Path]:
    """Resolve a path to a state file for the given filesystem.

    Handles params.yaml, zntrack.json, etc.

    Args:
        fs: The filesystem object
        state_path: The state path of the node
        filename: The filename to resolve (e.g., "params.yaml", "zntrack.json")
        validate_exists: Whether to validate that the path exists before returning

    Returns:
        Resolved path suitable for the filesystem

    Raises:
        FileNotFoundError: If validate_exists is True and the resolved path doesn't exist
    """
    target_path = state_path / filename
    resolved_path = resolve_dvc_path(fs, state_path, target_path)

    # Optionally validate that the path exists before returning it
    # This provides early error detection for missing files when requested
    if validate_exists and hasattr(fs, "exists") and not fs.exists(resolved_path):
        raise FileNotFoundError(f"State file not found: {resolved_path}")

    return resolved_path



import dataclasses
from pathlib import Path

import pytest
import yaml

import zntrack


@dataclasses.dataclass
class SimpleThermostat:
    """Simple thermostat class"""


@dataclasses.dataclass
class ThermostatA:
    temperature: float


@dataclasses.dataclass
class ThermostatB:
    temperature: float


@dataclasses.dataclass
class ClassWithDepsAndParams:
    """Class with deps and params"""

    temperature: float

    config: list[Path | str] | Path | str = zntrack.params_path()
    files: list[Path | str] | Path | str = zntrack.deps_path()


class HasBase:
    """Base class for unsupported fields."""


@dataclasses.dataclass
class HasOuts(HasBase):
    val: str = zntrack.outs()


@dataclasses.dataclass
class HasMetrics(HasBase):
    val: str = zntrack.metrics()


@dataclasses.dataclass
class HasPlots(HasBase):
    val: str = zntrack.plots()


@dataclasses.dataclass
class HasDeps(HasBase):
    val: str = zntrack.deps()


@dataclasses.dataclass
class HasParams(HasBase):
    val: str = zntrack.params()


class HasIllegalDC(zntrack.Node):
    method: HasBase = zntrack.deps()


class MD(zntrack.Node):
    thermostat: ThermostatA | ThermostatB | SimpleThermostat | ClassWithDepsAndParams = (
        zntrack.deps()
    )

    result: str = zntrack.outs()

    def run(self):
        self.result = self.thermostat.__class__.__name__


def test_switch_deps_class_keep_params(proj_path):
    """Test that changing the class triggers a DVC run."""
    proj = zntrack.Project()

    thermostat_a = ThermostatA(temperature=10)
    thermostat_b = ThermostatB(temperature=5)
    thermostat_c = SimpleThermostat()

    with proj:
        md = MD(thermostat=thermostat_a)

    proj.repro()
    assert md.from_rev().result == "ThermostatA"

    with proj:
        md.thermostat = thermostat_b

    proj.repro()
    assert md.from_rev().result == "ThermostatB"

    with proj:
        md.thermostat = thermostat_c

    proj.repro()
    assert md.from_rev().result == "SimpleThermostat"


def test_dc_deps_params_files(proj_path):
    config_file = Path("config.yaml")
    config_file.write_text(yaml.dump({"value": 10}))
    deps_file = Path("file.txt")
    deps_file.write_text("test")
    project = zntrack.Project()

    instance = ClassWithDepsAndParams(
        temperature=10,
        config=[config_file],
        files=deps_file,
    )
    with project:
        md = MD(thermostat=instance)

    project.repro()

    node = md.from_rev()
    assert isinstance(node.thermostat, ClassWithDepsAndParams)
    assert node.thermostat.temperature == 10
    assert node.thermostat.config == [config_file]
    assert node.thermostat.files == deps_file
    assert node.result == "ClassWithDepsAndParams"


@pytest.mark.parametrize(
    "node_class",
    [
        HasDeps,
        HasParams,
    ],
)
def test_dc_with_wrong_field_init(node_class):
    """Test that dataclass with wrong field raises an error."""
    project = zntrack.Project()
    instance = node_class(val="test")
    with project:
        HasIllegalDC(method=instance)
    with pytest.raises(TypeError):
        project.build()


@pytest.mark.parametrize(
    "node_class",
    [
        HasOuts,
        HasMetrics,
        HasPlots,
    ],
)
def test_dc_with_wrong_field_no_init(node_class):
    """Test that dataclass with wrong field raises an error."""
    project = zntrack.Project()
    instance = node_class()
    with project:
        HasIllegalDC(method=instance)
    with pytest.raises(TypeError):
        project.build()


if __name__ == "__main__":
    test_switch_deps_class_keep_params("")



import json
from pathlib import Path, PurePosixPath

import dvc.api
import pandas as pd
from dvc.stage import PipelineStage, Stage
from rich.console import Console
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn
from rich.text import Text
from rich.tree import Tree

from zntrack.group import Group
from zntrack.utils.state import get_node_status


def normalize_path(path: str) -> PurePosixPath:
    """Clean up path, removing '..' and '.'."""
    parts = [p for p in PurePosixPath(path).parts if p not in ("..", ".")]
    return PurePosixPath(*parts)


def format_node(short_name: str, full_name: str, changed: bool) -> Text:
    """Format tree leaf node."""
    if changed is True:
        status_symbol = "❌"
        style = "bold red"
    elif changed is False:
        status_symbol = "✅"
        style = "bold green"
    else:  # changed is None (unknown status)
        status_symbol = "❓"
        style = "bold blue"

    text = Text(f"{short_name} {status_symbol}", style=style)
    if short_name != full_name:
        text.append(f" -> {full_name}", style="dim")
    return text


def build_forest(df: pd.DataFrame) -> list[Tree]:
    """Build nested tree from dataframe grouped by 'group' column."""
    forest = []
    grouped = df.groupby(df["group"])
    trees_by_path = {}

    for group_path, items in grouped:
        group_path = tuple(group_path)

        for i, part in enumerate(group_path):
            path = group_path[: i + 1]
            if path not in trees_by_path:
                node_name = "📁 No Group" if part == "__NO_GROUP__" else f"📁 {part}"
                tree = Tree(node_name)
                trees_by_path[path] = tree
                if i == 0:
                    forest.append(tree)
                else:
                    trees_by_path[group_path[:i]].add(tree)

        for _, row in items.iterrows():
            trees_by_path[group_path].add(
                format_node(row["name"], row["full_name"], row["changed"])
            )

    return forest


def list_nodes(
    remote: str | None = None, rev: str | None = None, verbose: int = 1
) -> pd.DataFrame:
    """List zntrack nodes from DVC repo and display a nested tree."""
    fs = dvc.api.DVCFileSystem(url=remote, rev=rev)
    stages: list[Stage | PipelineStage] = list(fs.repo.stage.collect())
    node_data = []

    with Progress(
        SpinnerColumn(),
        # First "column": Overall progress count
        TextColumn("[cyan]Node: {task.completed}/{task.total} |"),
        BarColumn(),
        # Second "column": Current item being processed
        TextColumn(
            "[progress.description] [bold]{task.fields[current_node_address]}[/bold]"
        ),
        transient=True,  # Keeps the bar from lingering after completion
        refresh_per_second=10,
    ) as progress:
        task = progress.add_task(
            "[cyan]Checking nodes...",
            total=len(stages),
            current_node_address="",  # Initialize custom field
        )
        for idx, stage in enumerate(stages, 1):
            # Update the custom field for current node's addressing
            progress.update(task, completed=idx, current_node_address=stage.addressing)
            if not isinstance(stage, PipelineStage):
                continue
            # Determine stage_name and dvc_path if possible
            if ":" in stage.addressing:
                dvc_path_str, stage_name = stage.addressing.split(":")
                # Exclude 'dvc.yaml' from the dvc_parts for grouping
                dvc_path_obj = normalize_path(dvc_path_str)
                dvc_parts = tuple(p for p in dvc_path_obj.parts if p != "dvc.yaml")
            else:
                # Single file project
                stage_name = stage.addressing
                # If addressing is a dvc.yaml, we treat its parent as the dvc_parts
                if stage_name == "dvc.yaml":
                    dvc_parts = ()  # no dvc_parts if it's just dvc.yaml in root
                elif PurePosixPath(stage_name).name == "dvc.yaml":
                    dvc_parts = PurePosixPath(stage_name).parent.parts
                else:
                    dvc_parts = ()

            short_name = stage.name

            # Load zntrack group per node (from its nwd)
            try:
                config_path = Path(stage.path_in_repo).parent / "zntrack.json"
                config = json.loads(fs.read_text(config_path))
                nwd = config[stage.name]["nwd"]["value"]
                group = Group.from_nwd(Path(nwd))
                group_parts = tuple(group.names) if group.names else ()
            except Exception:
                group_parts = ()

            # Build group path
            if dvc_parts and group_parts:
                group_path = dvc_parts + group_parts
            elif dvc_parts:
                group_path = dvc_parts
            elif group_parts:
                group_path = group_parts
            else:
                group_path = ("__NO_GROUP__",)

            node_data.append(
                {
                    "name": short_name,
                    "full_name": stage.addressing,
                    "group": group_path,
                    "changed": get_node_status(stage.addressing, remote, rev, fs=fs),
                }
            )

    df = pd.DataFrame(node_data)

    # Render tree
    if verbose > 0:
        console = Console()
        for top_tree in build_forest(df):
            console.print(top_tree)

    return df



import dataclasses
import json
import pathlib
from pathlib import Path

import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


class Model:
    """Base Model class"""


@dataclasses.dataclass
class ModelWithParamsPath(Model):
    """Model with parameters"""

    params: dict
    config: str | Path | list[str | Path] = zntrack.params_path()


@dataclasses.dataclass
class ModelWithDepsPath(Model):
    """Model with dependencies"""

    params: dict
    files: str | Path | list[str | Path] = zntrack.deps_path()


@dataclasses.dataclass
class ModelWithParamsAndDepsPath(Model):
    """Model with parameters and dependencies"""

    params: dict
    config: str | Path | list[str | Path] = zntrack.params_path()
    files: str | Path | list[str | Path] = zntrack.deps_path()


class NodeWithModel(zntrack.Node):
    """Node with model"""

    model: Model | list[Model] = zntrack.deps()


def test_node_with_dc_model_params_deps(proj_path):
    project = zntrack.Project()

    a1 = ModelWithParamsPath(params={"a": 1}, config="config.yaml")
    a2 = ModelWithParamsPath(params={"a": 1}, config=Path("config.yaml"))
    b1 = ModelWithDepsPath(params={"a": 1}, files="file.txt")
    b2 = ModelWithDepsPath(params={"a": 1}, files=Path("file.txt"))
    c = ModelWithParamsAndDepsPath(
        params={"a": 1},
        config=["config.yaml", Path("config2.yaml")],
        files=["file.txt", Path("file2.txt")],
    )

    with project:
        NodeWithModel(model=a1)
        NodeWithModel(model=a2)
        NodeWithModel(model=b1)
        NodeWithModel(model=b2)
        NodeWithModel(model=c)
        NodeWithModel(model=[c])

    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "dataclass_w_deps_params_path.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "dataclass_w_deps_params_path.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "dataclass_w_deps_params_path.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    from pathlib import Path

    test_node_with_dc_model_params_deps(Path.cwd())



import dataclasses

import znflow

from zntrack.config import (
    FIELD_TYPE,
    FieldTypes,
)
from zntrack.node import Node
from zntrack.utils import module_handler


def _dataclass_to_dict(object) -> dict:
    """Convert a dataclass to a dictionary excluding certain keys."""
    exclude_fields = []
    for field in dataclasses.fields(object):
        if FIELD_TYPE in field.metadata:
            if field.metadata[FIELD_TYPE] in [
                FieldTypes.PARAMS_PATH,
                FieldTypes.DEPS_PATH,
                # FieldTypes.DEPS,
            ]:
                exclude_fields.append(field.name)
            else:
                raise TypeError(
                    f"Unsupported field type '{field.metadata[FIELD_TYPE]}'"
                    f" for field '{field.name}'."
                )
    dc_params = dataclasses.asdict(object)
    for f in exclude_fields:
        dc_params.pop(f)
    dc_params["_cls"] = f"{module_handler(object)}.{object.__class__.__name__}"
    return dc_params


def deps_to_params(self, field):
    if getattr(self.node, field.name) is None:
        return
    content = getattr(self.node, field.name)
    if isinstance(content, (list, tuple, dict)):
        new_content = []
        for val in content if isinstance(content, (list, tuple)) else content.values():
            if dataclasses.is_dataclass(val) and not isinstance(
                val, (Node, znflow.Connection, znflow.CombinedConnections)
            ):
                # We save the values of the passed dataclasses
                #  to the params.yaml file to be later used
                #  by the DataclassContainer to recreate the
                #  instance with the correct parameters.
                new_content.append(_dataclass_to_dict(val))
            elif isinstance(val, (znflow.Connection, znflow.CombinedConnections)):
                pass
            else:
                raise ValueError(
                    f"Found unsupported type '{type(val)}' ({val}) for DEPS"
                    f" field '{field.name}' in list"
                )
        if len(new_content) > 0:
            return new_content
    elif dataclasses.is_dataclass(content) and not isinstance(
        content, (Node, znflow.Connection, znflow.CombinedConnections)
    ):
        return _dataclass_to_dict(content)
    elif isinstance(content, (znflow.Connection, znflow.CombinedConnections)):
        return
    else:
        raise ValueError(
            f"Found unsupported type '{type(content)}' ({content})"
            f" for DEPS field '{field.name}'"
        )



import json
from pathlib import Path
from unittest.mock import MagicMock

import pytest

import zntrack
from zntrack.utils.lockfile import get_stage_lock


class ReadFileContent(zntrack.Node):
    deps_file: Path = zntrack.deps_path()
    params: str = zntrack.params()

    def run(self):
        pass


@pytest.fixture()
def lockfile_01():
    return {
        "cmd": "zntrack run test_node_meta.ReadFileContent --name ReadFileContent",
        "deps": [
            {
                "hash": "md5",
                "md5": "6dbd01b4309de2c22b027eb35a3ce18b",
                "path": "data.txt",
                "size": 11,
            }
        ],
        "params": {"params.yaml": {"ReadFileContent": {"params": "test"}}},
    }


def test_node_meta_lock(proj_path, lockfile_01):
    project = zntrack.Project()

    file = Path("data.txt")
    file.write_text("Lorem Ipsum")

    with project:
        _ = ReadFileContent(deps_file=file, params="test")

    project.repro()
    # TODO: do we want the node to require from_rev()?
    #   to update lockfile and other node-meta states?
    node = ReadFileContent.from_rev()
    assert node.state.lockfile == lockfile_01


def test_node_meta_lock_mp(proj_path, lockfile_01):
    project = zntrack.Project()

    file = Path("data.txt")
    file.write_text("Lorem Ipsum")

    with project:
        node = ReadFileContent(deps_file=file, params="test")

    project.repro()

    queue = MagicMock()  # mock multiprocessing.Queue
    get_stage_lock(node.name, queue)
    assert queue.put.call_count == 1
    lockfile = json.loads(json.dumps(queue.put.call_args[0][0]))
    assert lockfile["cmd"] == lockfile_01["cmd"]
    assert lockfile["deps"] == lockfile_01["deps"]
    assert lockfile["params"] == lockfile_01["params"]



import json
import subprocess

import pytest
from typer.testing import CliRunner

import zntrack
import zntrack.examples
from zntrack.cli import app


@pytest.fixture()
def runner() -> CliRunner:
    return CliRunner()


def test_version(runner):
    result = runner.invoke(app, ["--version"])
    assert "ZnTrack " in result.stdout  # ZnTrack v.x.y


def test_run(proj_path, runner):
    with zntrack.Project() as proj:
        node = zntrack.examples.ParamsToOuts(params=15)

    proj.build()
    result = runner.invoke(app, ["run", "zntrack.examples.ParamsToOuts"])
    assert result.exit_code == 0

    assert node.outs == 15


def test_run_w_name(proj_path, runner):
    with zntrack.Project() as proj:
        node = zntrack.examples.ParamsToOuts(params=15, name="TestNode")

    proj.build()

    result = runner.invoke(
        app, ["run", "zntrack.examples.ParamsToOuts", "--name", "TestNode"]
    )
    assert result.exit_code == 0

    assert node.outs == 15


def test_list_groups(proj_path, runner):
    with zntrack.Project() as proj:
        a = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    with proj.group("example1"):
        b = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    with proj.group("nested", "GRP1"):
        c = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)
    with proj.group("nested", "GRP2"):
        d = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    proj.build()

    subprocess.check_call(["dvc", "repro", a.name, b.name, c.name, d.name])

    true_groups = [
        {
            "changed": False,
            "full_name": "ParamsToOuts",
            "group": [
                "__NO_GROUP__",
            ],
            "name": "ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "ParamsToOuts_1",
            "group": [
                "__NO_GROUP__",
            ],
            "name": "ParamsToOuts_1",
        },
        {
            "changed": False,
            "full_name": "example1_ParamsToOuts",
            "group": [
                "example1",
            ],
            "name": "example1_ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "example1_ParamsToOuts_1",
            "group": [
                "example1",
            ],
            "name": "example1_ParamsToOuts_1",
        },
        {
            "changed": False,
            "full_name": "nested_GRP1_ParamsToOuts",
            "group": [
                "nested",
                "GRP1",
            ],
            "name": "nested_GRP1_ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "nested_GRP1_ParamsToOuts_1",
            "group": [
                "nested",
                "GRP1",
            ],
            "name": "nested_GRP1_ParamsToOuts_1",
        },
        {
            "changed": False,
            "full_name": "nested_GRP2_ParamsToOuts",
            "group": [
                "nested",
                "GRP2",
            ],
            "name": "nested_GRP2_ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "nested_GRP2_ParamsToOuts_1",
            "group": [
                "nested",
                "GRP2",
            ],
            "name": "nested_GRP2_ParamsToOuts_1",
        },
    ]

    result = runner.invoke(app, ["list", proj_path.as_posix(), "--json"])
    groups = json.loads(result.stdout)
    assert groups == true_groups

    assert result.exit_code == 0

    result = runner.invoke(app, ["list", proj_path.as_posix()])
    assert result.exit_code == 0

    outs = """
📁 No Group
├── ParamsToOuts ✅
└── ParamsToOuts_1 ❌
📁 example1
├── example1_ParamsToOuts ✅
└── example1_ParamsToOuts_1 ❌
📁 nested
├── 📁 GRP1
│   ├── nested_GRP1_ParamsToOuts ✅
│   └── nested_GRP1_ParamsToOuts_1 ❌
└── 📁 GRP2
    ├── nested_GRP2_ParamsToOuts ✅
    └── nested_GRP2_ParamsToOuts_1 ❌
"""
    assert result.stdout in outs


def test_list_multi_nested_groups(proj_path, runner):
    proj = zntrack.Project()

    with proj:
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    with proj.group("dynamics"):
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    with proj.group("dynamics", "400K"):
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    with proj.group("dynamics", "400K", "B"):
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    proj.build()

    true_groups = [
        {
            "changed": True,
            "full_name": "ParamsToOuts",
            "group": [
                "__NO_GROUP__",
            ],
            "name": "ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "ParamsToOuts_1",
            "group": [
                "__NO_GROUP__",
            ],
            "name": "ParamsToOuts_1",
        },
        {
            "changed": True,
            "full_name": "dynamics_400K_B_ParamsToOuts",
            "group": [
                "dynamics",
                "400K",
                "B",
            ],
            "name": "dynamics_400K_B_ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "dynamics_400K_B_ParamsToOuts_1",
            "group": [
                "dynamics",
                "400K",
                "B",
            ],
            "name": "dynamics_400K_B_ParamsToOuts_1",
        },
        {
            "changed": True,
            "full_name": "dynamics_400K_ParamsToOuts",
            "group": [
                "dynamics",
                "400K",
            ],
            "name": "dynamics_400K_ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "dynamics_400K_ParamsToOuts_1",
            "group": [
                "dynamics",
                "400K",
            ],
            "name": "dynamics_400K_ParamsToOuts_1",
        },
        {
            "changed": True,
            "full_name": "dynamics_ParamsToOuts",
            "group": [
                "dynamics",
            ],
            "name": "dynamics_ParamsToOuts",
        },
        {
            "changed": True,
            "full_name": "dynamics_ParamsToOuts_1",
            "group": [
                "dynamics",
            ],
            "name": "dynamics_ParamsToOuts_1",
        },
    ]

    result = runner.invoke(app, ["list", proj_path.as_posix(), "--json"])
    groups = json.loads(result.stdout)
    assert groups == true_groups

    assert result.exit_code == 0

    result = runner.invoke(app, ["list", proj_path.as_posix()])
    assert result.exit_code == 0

    outs = """
📁 No Group
├── ParamsToOuts ❌
└── ParamsToOuts_1 ❌
📁 dynamics
├── dynamics_ParamsToOuts ❌
├── dynamics_ParamsToOuts_1 ❌
└── 📁 400K
    ├── dynamics_400K_ParamsToOuts ❌
    ├── dynamics_400K_ParamsToOuts_1 ❌
    └── 📁 B
        ├── dynamics_400K_B_ParamsToOuts ❌
        └── dynamics_400K_B_ParamsToOuts_1 ❌
"""
    assert result.stdout in outs
